@@ -372,7 +372,7 @@ void send_cmode(const char *sender, ...)
 		flush_cmode(&modedata[which]);
 		md->used = time(NULL);
 	    }
-	    s = md->opmodes;
+	    s = md->opmodes + strlen(md->opmodes);
 	    if (add != md->last_add) {
 		*s++ = add ? '+' : '-';
 		md->last_add = add;
@@ -386,6 +386,7 @@ void send_cmode(const char *sender, ...)
 			 "%s%s", md->paramslen ? " " : "", s);
 	    md->nparams++;
 	    break;
+
 	  default:
 	    flag = mode_char_to_flag(c, MODE_CHANNEL);
 	    if (add) {
@@ -424,20 +425,22 @@ static void flush_cmode(struct modedata *md)
     if (!md->binmodes_on && !md->binmodes_off && !*md->opmodes)
 	goto done;	/* No actual modes here */
 
+    /* Note that - must come before + because some servers (Unreal, others?)
+     * ignore +s if followed by -p. */
+    if (md->binmodes_off) {
+	len += snprintf(buf+len, sizeof(buf)-len, "-%s",
+			mode_flags_to_string(md->binmodes_off, MODE_CHANNEL));
+	lastc = '-';
+    }
     if (md->binmodes_on) {
 	len += snprintf(buf+len, sizeof(buf)-len, "+%s",
 			mode_flags_to_string(md->binmodes_on, MODE_CHANNEL));
 	lastc = '+';
     }
-    if (md->binmodes_off) {
-	len += snprintf(buf+len, sizeof(buf)-len, "-%s",
-			mode_flags_to_string(md->binmodes_on, MODE_CHANNEL));
-	lastc = '-';
-    }
     if (*md->opmodes) {
 	if (*md->opmodes == lastc) {
 	    /* First +/- matches last binary mode change */
-	    memmove(md->opmodes+1, md->opmodes, strlen(md->opmodes+1)+1);
+	    memmove(md->opmodes, md->opmodes+1, strlen(md->opmodes+1)+1);
 	}
 	len += snprintf(buf+len, sizeof(buf)-len, "%s", md->opmodes);
     }
