@@ -93,9 +93,11 @@ static int def_levels[][2] = {
     { CA_ACCESS_LIST,        0 },
     { CA_ACCESS_CHANGE,     10 },
     { CA_MEMO,              10 },
-    { CA_VOICEDEVOICE,       3 },
+    { CA_VOICE,              3 },
     { CA_AUTOHALFOP,         4 },
-    { CA_HALFOPDEHALFOP,     4 },
+    { CA_HALFOP,             4 },
+    { CA_AUTOPROTECT,       10 },
+    { CA_PROTECT,           10 },
     { -1 }
 };
 
@@ -124,10 +126,14 @@ static LevelInfo levelinfo[] = {
     { CA_ACCESS_LIST,   "ACC-LIST",   CHAN_LEVEL_ACCESS_LIST },
     { CA_ACCESS_CHANGE, "ACC-CHANGE", CHAN_LEVEL_ACCESS_CHANGE },
     { CA_MEMO,          "MEMO",       CHAN_LEVEL_MEMO },
-    { CA_VOICEDEVOICE,  "VOICE-DEVOICE", CHAN_LEVEL_VOICEDEVOICE },
+    { CA_VOICE,         "VOICE",      CHAN_LEVEL_VOICE },
 #ifdef HAVE_HALFOP
     { CA_AUTOHALFOP,    "AUTOHALFOP", CHAN_LEVEL_AUTOHALFOP },
-    { CA_HALFOPDEHALFOP,"HALFOP-DEHALFOP", CHAN_LEVEL_HALFOPDEHALFOP },
+    { CA_HALFOP,        "HALFOP",     CHAN_LEVEL_HALFOP },
+#endif
+#ifdef HAVE_CHANPROT
+    { CA_AUTOPROTECT,   "AUTOPROTECT",CHAN_LEVEL_AUTOPROTECT },
+    { CA_PROTECT,       "PROTECT",    CHAN_LEVEL_PROTECT },
 #endif
     { -1 }
 };
@@ -149,6 +155,7 @@ static int get_access(User *user, ChannelInfo *ci);
 static void suspend(ChannelInfo *ci, const char *reason,
 		    const char *who, const time_t expires);
 static void unsuspend(ChannelInfo *ci, int set_time);
+static void chan_bad_password(User *u, ChannelInfo *ci);
 static ChanOpt *chanopt_from_name(const char *optname);
 
 static void do_help(User *u);
@@ -187,6 +194,10 @@ static void do_devoice(User *u);
 static void do_halfop(User *u);
 static void do_dehalfop(User *u);
 #endif
+#ifdef HAVE_CHANPROT
+static void do_protect(User *u);
+static void do_deprotect(User *u);
+#endif
 static void do_unban(User *u);
 static void do_clear(User *u);
 static void do_getpass(User *u);
@@ -263,6 +274,10 @@ static Command cmds[] = {
 #ifdef HAVE_HALFOP
     { "HALFOP",   do_halfop,   NULL,  CHAN_HELP_HALFOP,         -1,-1,-1,-1 },
     { "DEHALFOP", do_dehalfop, NULL,  CHAN_HELP_DEHALFOP,       -1,-1,-1,-1 },
+#endif
+#ifdef HAVE_CHANPROT
+    { "PROTECT",  do_protect,  NULL,  CHAN_HELP_PROTECT,        -1,-1,-1,-1 },
+    { "DEPROTECT",do_deprotect,NULL,  CHAN_HELP_DEPROTECT,      -1,-1,-1,-1 },
 #endif
     { "INVITE",   do_invite,   NULL,  CHAN_HELP_INVITE,         -1,-1,-1,-1 },
     { "UNBAN",    do_unban,    NULL,  CHAN_HELP_UNBAN,          -1,-1,-1,-1 },
@@ -317,7 +332,7 @@ void chanserv(const char *source, char *buf)
 
     if (!u) {
 	log("%s: user record for %s not found", s_ChanServ, source);
-	notice(s_NickServ, source,
+	notice(s_ChanServ, source,
 		getstring((NickInfo *)NULL, USER_RECORD_NOT_FOUND));
 	return;
     }
@@ -579,16 +594,35 @@ int check_chan_user_modes(const char *source, User *user, const char *chan,
 	return modes;
 #endif
 
+    /* Check early for server auto-ops */
+    if ((modes & CUMODE_o)
+     && !(ci->flags & CI_LEAVEOPS)
+     && is_servermode
+     && time(NULL)-start_time >= CSRestrictDelay
+     && !check_access(user, ci, CA_AUTOOP)
+    ) {
+	notice_lang(s_ChanServ, user, CHAN_IS_REGISTERED, s_ChanServ);
+	send_cmode(MODE_SENDER(s_ChanServ), chan, "-o", user->nick);
+	modes &= ~CUMODE_o;
+    }
+
     /* Mode additions.  Only check for join or server mode change, unless
      * ENFORCE is set */
     if (is_servermode || (ci->flags & CI_ENFORCE)) {
-	int set = 1;
-#ifdef IRC_UNREAL
+	int set = 0;
+#ifdef HAVE_CHANPROT
 	if (is_founder(user, ci)) {
 	    ci->last_used = time(NULL);
 	    if (!(modes & CUMODE_q))
 		send_cmode(MODE_SENDER(s_ChanServ), chan, "+q", user->nick);
 	    modes |= CUMODE_q;
+	    set++;
+	} else if (check_access(user, ci, CA_AUTOPROTECT)) {
+	    ci->last_used = time(NULL);
+	    if (!(modes & CUMODE_a))
+		send_cmode(MODE_SENDER(s_ChanServ), chan, "+a", user->nick);
+	    modes |= CUMODE_a;
+	    set++;
 	}
 #endif
 	if (check_access(user, ci, CA_AUTOOP)) {
@@ -596,18 +630,19 @@ int check_chan_user_modes(const char *source, User *user, const char *chan,
 	    if (!(modes & CUMODE_o))
 		send_cmode(MODE_SENDER(s_ChanServ), chan, "+o", user->nick);
 	    modes |= CUMODE_o;
+	    set++;
 #ifdef HAVE_HALFOP
 	} else if (check_access(user, ci, CA_AUTOHALFOP)) {
 	    if (!(modes & CUMODE_h))
 		send_cmode(MODE_SENDER(s_ChanServ), chan, "+h", user->nick);
 	    modes |= CUMODE_h;
+	    set++;
 #endif
 	} else if (check_access(user, ci, CA_AUTOVOICE)) {
 	    if (!(modes & CUMODE_v))
 		send_cmode(MODE_SENDER(s_ChanServ), chan, "+v", user->nick);
 	    modes |= CUMODE_v;
-	} else {
-	    set = 0;
+	    set++;
 	}
 	if (set)
 	    return modes;  /* don't bother checking for mode subtractions */
@@ -618,27 +653,28 @@ int check_chan_user_modes(const char *source, User *user, const char *chan,
 	return modes;
 
     /* Mode subtractions */
-    if (modes & CUMODE_o) {
-	if (!(ci->flags & CI_LEAVEOPS) && is_servermode
-	    && time(NULL)-start_time >= CSRestrictDelay
-	    && !check_access(user, ci, CA_AUTOOP)
-	) {
-	    notice_lang(s_ChanServ, user, CHAN_IS_REGISTERED, s_ChanServ);
-	    send_cmode(MODE_SENDER(s_ChanServ), chan, "-o", user->nick);
-	    modes &= ~CUMODE_o;
-	} else if (check_access(user, ci, CA_AUTODEOP)) {
+    if (check_access(user, ci, CA_AUTODEOP)) {
+	if (modes & CUMODE_o) {
 	    send_cmode(MODE_SENDER(s_ChanServ), chan, "-o", user->nick);
 	    modes &= ~CUMODE_o;
 	}
-    }
-#ifdef HAVE_HALFOPS
-    if (modes & CUMODE_h) {
-	if (check_access(user, ci, CA_AUTODEOP)) {
+#ifdef HAVE_HALFOP
+	if (modes & CUMODE_h) {
 	    send_cmode(MODE_SENDER(s_ChanServ), chan, "-h", user->nick);
 	    modes &= ~CUMODE_h;
 	}
-    }
 #endif
+#ifdef HAVE_HALFOP
+	if (modes & CUMODE_a) {
+	    send_cmode(MODE_SENDER(s_ChanServ), chan, "-a", user->nick);
+	    modes &= ~CUMODE_a;
+	}
+	if (modes & CUMODE_q) {
+	    send_cmode(MODE_SENDER(s_ChanServ), chan, "-q", user->nick);
+	    modes &= ~CUMODE_q;
+	}
+#endif
+    }
 
     return modes;
 }
@@ -682,7 +718,13 @@ int check_kick(User *user, const char *chan)
 #ifdef IRC_UNREAL
     /* Don't let plain opers into +A (admin only) channels */
     if ((ci->mlock_on & CMODE_A) && !(user->mode & (UMODE_A|UMODE_N|UMODE_T))){
-	mask = create_mask(user);
+	mask = create_mask(user, 1);
+	reason = getstring(user->ni, CHAN_NOT_ALLOWED_TO_JOIN);
+	goto kick;
+    }
+    /* Don't let hiding users into no-hiding channels */
+    if ((ci->mlock_on & CMODE_H) && (user->mode & UMODE_I)){
+	mask = create_mask(user, 1);
 	reason = getstring(user->ni, CHAN_NOT_ALLOWED_TO_JOIN);
 	goto kick;
     }
@@ -699,7 +741,7 @@ int check_kick(User *user, const char *chan)
 
     if (ci->mlock_on & CMODE_OPERONLY) {
 	/* We already know they're not an oper, so kick them off */
-	mask = create_mask(user);
+	mask = create_mask(user, 1);
 	reason = getstring(user->ni, CHAN_NOT_ALLOWED_TO_JOIN);
 	goto kick;
     }
@@ -713,13 +755,14 @@ int check_kick(User *user, const char *chan)
 	if (!akick->in_use)
 	    continue;
 	if ((akick->is_nick && getlink(akick->u.ni) == ni)
-		|| (!akick->is_nick && match_usermask(akick->u.mask, user))
+	 || (!akick->is_nick && match_usermask(akick->u.mask, user))
 	) {
 	    if (debug >= 2) {
 		log("debug: %s matched akick %s", user->nick,
 			akick->is_nick ? akick->u.ni->nick : akick->u.mask);
 	    }
-	    mask = akick->is_nick ? create_mask(user) : sstrdup(akick->u.mask);
+	    mask = akick->is_nick ? create_mask(user, 1)
+	                          : sstrdup(akick->u.mask);
 	    reason = akick->reason ? akick->reason : CSAutokickReason;
 	    goto kick;
 	}
@@ -727,7 +770,7 @@ int check_kick(User *user, const char *chan)
 
     if (time(NULL)-start_time >= CSRestrictDelay
 				&& check_access(user, ci, CA_NOJOIN)) {
-	mask = create_mask(user);
+	mask = create_mask(user, 1);
 	reason = getstring(user->ni, CHAN_NOT_ALLOWED_TO_JOIN);
 	goto kick;
     }
@@ -871,26 +914,38 @@ void cs_remove_nick(const NickInfo *ni)
     for (ci = cs_firstchan(); ci; ci = next) {
 	next = cs_nextchan();
 	if (ci->founder == ni) {
+	    int was_suspended = (ci->suspendinfo != NULL);
+	    char name_save[CHANMAX];
+	    if (was_suspended)
+		strscpy(name_save, ci->name, CHANMAX);
+	    uncount_chan(ci);  /* Make sure it disappears from founderchans */
 	    if (ci->successor) {
 		NickInfo *ni2 = ci->successor;
-		if (check_channel_limit(ni2) >= 0) {
-		    log("%s: Successor (%s) of %s owns too many channels, "
-			"deleting channel",
-			s_ChanServ, ni2->nick, ci->name);
-		    delchan(ci);
-		    continue;
-		} else {
+		if (check_channel_limit(ni2) < 0) {
 		    log("%s: Transferring foundership of %s from deleted "
 			"nick %s to successor %s",
 			s_ChanServ, ci->name, ni->nick, ni2->nick);
 		    ci->founder = ni2;
 		    ci->successor = NULL;
 		    count_chan(ci);
+		} else {
+		    log("%s: Successor (%s) of %s owns too many channels, "
+			"deleting channel",
+			s_ChanServ, ni2->nick, ci->name);
+		    goto delete;
 		}
 	    } else {
 		log("%s: Deleting channel %s owned by deleted nick %s",
 		    s_ChanServ, ci->name, ni->nick);
+	      delete:
 		delchan(ci);
+		if (was_suspended) {
+		    /* Channel was suspended, so make it forbidden */
+		    log("%s: Channel %s was suspended, forbidding it",
+			s_ChanServ, name_save);
+		    ci = makechan(name_save);
+		    ci->flags |= CI_VERBOTEN;
+		}
 		continue;
 	    }
 	}
@@ -962,8 +1017,7 @@ int check_channel_limit(NickInfo *ni)
 /*************************************************************************/
 
 /* Return a string listing the options (those given in chanopts[]) set on
- * the given channel.  Uses the given NickInfo for language information;
- * if NULL, the option names themselves ("KEEPTOPIC", etc.) are used.
+ * the given channel.  Uses the given NickInfo for language information.
  * The returned string is stored in a static buffer which will be
  * overwritten on the next call.
  */
@@ -972,7 +1026,7 @@ char *chanopts_to_string(ChannelInfo *ci, NickInfo *ni)
 {
     static char buf[BUFSIZE];
     char *end = buf;
-    const char *commastr = ni ? getstring(ni, COMMA_SPACE) : ", ";
+    const char *commastr = getstring(ni, COMMA_SPACE);
     const char *s;
     int need_comma = 0;
     int i;
@@ -980,8 +1034,8 @@ char *chanopts_to_string(ChannelInfo *ci, NickInfo *ni)
     for (i = 0; chanopts[i].name && end-buf < sizeof(buf)-1; i++) {
 	if (ni && chanopts[i].namestr < 0)
 	    continue;
-	if (ci->flags & chanopts[i].flag) {
-	    s = ni ? getstring(ni, chanopts[i].namestr) : chanopts[i].name;
+	if (ci->flags & chanopts[i].flag && chanopts[i].namestr >= 0) {
+	    s = getstring(ni, chanopts[i].namestr);
 	    if (need_comma)
 		end += snprintf(end, sizeof(buf)-(end-buf), "%s", commastr);
 	    end += snprintf(end, sizeof(buf)-(end-buf), "%s", s);
@@ -1043,7 +1097,29 @@ static ChannelInfo *makechan(const char *chan)
 static int delchan(ChannelInfo *ci)
 {
     int i;
+    User *u;
 
+    /* Remove channel from founder's owned-channel count */
+    uncount_chan(ci);
+
+    /* Clear channel from users' identified-channel lists */
+    for (u = firstuser(); u; u = nextuser()) {
+	struct u_chaninfolist *uc, *next;
+	for (uc = u->founder_chans; uc; uc = next) {
+	    next = uc->next;
+	    if (uc->chan == ci) {
+		if (uc->next)
+		    uc->next->prev = uc->prev;
+		if (uc->prev)
+		    uc->prev->next = uc->next;
+		else
+		    u->founder_chans = uc->next;
+		free(uc);
+	    }
+	}
+    }
+
+    /* Now actually free channel data */
     if (ci->c)
 	ci->c->ci = NULL;
     if (ci->next)
@@ -1079,8 +1155,8 @@ static int delchan(ChannelInfo *ci)
 	}
 	free(ci->memos.memos);
     }
-    uncount_chan(ci);
     free(ci);
+
     return 1;
 }
 
@@ -1125,7 +1201,7 @@ static void uncount_chan(ChannelInfo *ci)
 	if (ni->founderchans[i] == ci) {
 	    ni->foundercount--;
 	    if (i < ni->foundercount)
-		memmove(&ni->founderchans[i+1], &ni->founderchans[i],
+		memmove(&ni->founderchans[i], &ni->founderchans[i+1],
 			sizeof(*ni->founderchans) * (ni->foundercount-i));
 	    ni->founderchans = srealloc(ni->founderchans,
 			sizeof(*ni->founderchans) * ni->foundercount);
@@ -1264,6 +1340,31 @@ static void unsuspend(ChannelInfo *ci, int set_time)
 
 /*************************************************************************/
 
+/* Register a bad password attempt for a channel. */
+
+static void chan_bad_password(User *u, ChannelInfo *ci)
+{
+    bad_password(s_ChanServ, u, ci->name);
+    ci->bad_passwords++;
+    if (BadPassWarning && ci->bad_passwords == BadPassWarning) {
+	wallops(s_ChanServ, "\2Warning:\2 Repeated bad password attempts"
+	                    " for channel %s", ci->name);
+    }
+    if (BadPassSuspend && ci->bad_passwords == BadPassSuspend) {
+	time_t expire = 0;
+	if (CSSuspendExpire)
+	    expire = time(NULL) + CSSuspendExpire;
+	suspend(ci, "Too many bad passwords (automatic suspend)",
+		s_ChanServ, expire);
+	log("%s: Automatic suspend for %s (too many bad passwords)",
+	    s_ChanServ, ci->name);
+	/* Clear bad password count for when channel is unsuspended */
+	ci->bad_passwords = 0;
+    }
+}
+
+/*************************************************************************/
+
 /* Return the ChanOpt structure for the given option name.  If not found,
  * return NULL.
  */
@@ -1288,8 +1389,12 @@ static void do_help(User *u)
 
     if (!cmd) {
 	notice_help(s_ChanServ, u, CHAN_HELP);
-#ifdef HAVE_HALFOP
+#if defined(HAVE_HALFOP) && defined(HAVE_CHANPROT)
+	notice_help(s_ChanServ, u, CHAN_HELP_OTHERS_HALFOP_CHANPROT);
+#elif defined(HAVE_HALFOP)
 	notice_help(s_ChanServ, u, CHAN_HELP_OTHERS_HALFOP);
+#elif defined(HAVE_CHANPROT)
+	notice_help(s_ChanServ, u, CHAN_HELP_OTHERS_CHANPROT);
 #else
 	notice_help(s_ChanServ, u, CHAN_HELP_OTHERS);
 #endif
@@ -1322,6 +1427,14 @@ static void do_help(User *u)
 	notice_help(s_ChanServ, u, CHAN_HELP_CLEAR_HALFOPS);
 #endif
 	notice_help(s_ChanServ, u, CHAN_HELP_CLEAR_END);
+    } else if (stricmp(cmd, "SOP") == 0) {
+	notice_help(s_ChanServ, u, CHAN_HELP_SOP);
+#ifdef HAVE_CHANPROT
+	notice_help(s_ChanServ, u, CHAN_HELP_SOP_PROT);
+#else
+	notice_help(s_ChanServ, u, CHAN_HELP_SOP_NOPROT);
+#endif
+	notice_help(s_ChanServ, u, CHAN_HELP_SOP_END);
     } else {
 	help_cmd(s_ChanServ, u, cmds, cmd);
     }
@@ -1461,6 +1574,8 @@ static void do_identify(User *u)
     } else {
 	int res = check_password(pass, ci->founderpass);
 	if (res == 1) {
+	    ci->bad_passwords = 0;
+	    ci->last_used = time(NULL);
 	    if (!is_identified(u, ci)) {
 		uc = smalloc(sizeof(*uc));
 		uc->next = u->founder_chans;
@@ -1479,7 +1594,7 @@ static void do_identify(User *u)
 	} else {
 	    log("%s: Failed IDENTIFY for %s by %s!%s@%s",
 			s_ChanServ, ci->name, u->nick, u->username, u->host);
-	    bad_password(s_ChanServ, u, ci->name);
+	    chan_bad_password(u, ci);
 	}
 
     }
@@ -1503,7 +1618,7 @@ static void do_drop(User *u)
 	syntax_error(s_ChanServ, u, "DROP", CHAN_DROP_SYNTAX);
     } else if (!(ci = cs_findchan(chan))) {
 	notice_lang(s_ChanServ, u, CHAN_X_NOT_REGISTERED, chan);
-    } else if (!is_servadmin & (ci->flags & CI_VERBOTEN)) {
+    } else if (!is_servadmin && (ci->flags & CI_VERBOTEN)) {
 	notice_lang(s_ChanServ, u, CHAN_X_FORBIDDEN, chan);
     } else if (!is_servadmin && !is_identified(u, ci)) {
 	notice_lang(s_ChanServ, u, CHAN_IDENTIFY_REQUIRED, s_ChanServ, chan);
@@ -1889,6 +2004,16 @@ static void do_set_mlock(User *u, ChannelInfo *ci, char *param)
 #endif
 	  default: {
 	    int32 flag = mode_char_to_flag(c, MODE_CHANNEL);
+#ifdef IRC_UNREAL
+	    if ((flag & (CMODE_A | CMODE_H))
+		&& !(is_services_admin(u)
+		     || (u->mode & (UMODE_A|UMODE_N|UMODE_T)))
+	    ) {
+		continue;
+	    }
+#endif
+	    if ((flag & CMODE_OPERONLY) && !is_oper_u(u))
+		continue;
 	    if (flag) {
 		if (flag & CMODE_REG)
 		    notice_lang(s_ChanServ, u, CHAN_SET_MLOCK_MODE_REG_BAD, c);
@@ -2224,8 +2349,12 @@ static void do_akick(User *u)
 	notice_lang(s_ChanServ, u, CHAN_AKICK_ENFORCE_DONE, chan, count);
 
     } else if (stricmp(cmd, "COUNT") == 0) {
-        notice_lang(s_ChanServ, u, CHAN_AKICK_COUNT, ci->name,
-                        ci->akickcount);
+	int count = 0, i;
+	for (i = 0; i < ci->akickcount; i++) {
+	    if (ci->akick[i].in_use)
+		count++;
+	}
+        notice_lang(s_ChanServ, u, CHAN_AKICK_COUNT, ci->name, count);
 
     } else {
 	syntax_error(s_ChanServ, u, "AKICK", CHAN_AKICK_SYNTAX);
@@ -2421,7 +2550,7 @@ static void do_info(User *u)
 	    notice_lang(s_ChanServ, u, CHAN_INFO_NO_EXPIRE);
 
 	if (ci->suspendinfo) {
-	    notice_lang(s_NickServ, u, CHAN_X_SUSPENDED, chan);
+	    notice_lang(s_ChanServ, u, CHAN_X_SUSPENDED, chan);
 	    if (CHECK_SHOW_ALL) {
 		SuspendInfo *si = ci->suspendinfo;
 		char timebuf[BUFSIZE], expirebuf[BUFSIZE];
@@ -2431,9 +2560,9 @@ static void do_info(User *u)
 			      STRFTIME_DATE_TIME_FORMAT, tm);
 		expires_in_lang(expirebuf, sizeof(expirebuf), u->ni,
 				si->expires);
-		notice_lang(s_NickServ, u, CHAN_INFO_SUSPEND_DETAILS,
+		notice_lang(s_ChanServ, u, CHAN_INFO_SUSPEND_DETAILS,
 			    si->who, timebuf, expirebuf);
-		notice_lang(s_NickServ, u, CHAN_INFO_SUSPEND_REASON,
+		notice_lang(s_ChanServ, u, CHAN_INFO_SUSPEND_REASON,
 			    si->reason);
 	    }
 	}
@@ -2540,7 +2669,7 @@ static void do_invite(User *u)
     } else if (!(c = findchan(chan))) {
 	notice_lang(s_ChanServ, u, CHAN_X_NOT_IN_USE, chan);
     } else if (c->bouncy_modes) {
-	notice_lang(s_ChanServ, u, CHAN_BOUNCY_MODES, chan, "INVITE");
+	notice_lang(s_ChanServ, u, CHAN_BOUNCY_MODES, "INVITE");
     } else if (!(ci = c->ci)) {
 	notice_lang(s_ChanServ, u, CHAN_X_NOT_REGISTERED, chan);
     } else if (ci->flags & CI_VERBOTEN) {
@@ -2561,25 +2690,34 @@ static struct {
     const char *cmd;
     int add;
     int32 mode;
-    int sender_acc, target_acc, target_nextacc, success_msg, failure_msg;
+    int sender_acc;	/* Access required for the sender */
+    int target_acc;	/* Target access level at which we refuse command */
+    int target_nextacc;	/* Level above target_acc (voice -> halfop etc) */
+    int success_msg, failure_msg;
 } opvoice_data[] = {
-    { "OP",       1, CUMODE_o, CA_OPDEOP, CA_AUTODEOP, -1,
+    { "OP",        1, CUMODE_o, CA_OPDEOP, CA_AUTODEOP, -1,
 	  CHAN_OP_SUCCEEDED, CHAN_OP_FAILED },
-    { "DEOP",     0, CUMODE_o, CA_OPDEOP, CA_AUTOOP, -1,
+    { "DEOP",      0, CUMODE_o, CA_OPDEOP, CA_AUTOOP, -1,
 	  CHAN_DEOP_SUCCEEDED, CHAN_DEOP_FAILED },
-    { "VOICE",    1, CUMODE_v, CA_VOICEDEVOICE, -1, -1,
+    { "VOICE",     1, CUMODE_v, CA_VOICE, -1, -1,
 	  CHAN_VOICE_SUCCEEDED, CHAN_VOICE_FAILED },
 #ifndef HAVE_HALFOP
-    { "DEVOICE",  0, CUMODE_v, CA_VOICEDEVOICE, CA_AUTOVOICE, CA_AUTOOP,
+    { "DEVOICE",   0, CUMODE_v, CA_VOICE, CA_AUTOVOICE, CA_AUTOOP,
 	  CHAN_DEVOICE_SUCCEEDED, CHAN_DEVOICE_FAILED },
-#else HAVE_HALFOP
-    { "DEVOICE",  0, CUMODE_v, CA_VOICEDEVOICE, CA_AUTOVOICE, CA_AUTOHALFOP,
+#else
+    { "DEVOICE",   0, CUMODE_v, CA_VOICE, CA_AUTOVOICE, CA_AUTOHALFOP,
 	  CHAN_DEVOICE_SUCCEEDED, CHAN_DEVOICE_FAILED },
-    { "HALFOP",   1, CUMODE_h, CA_HALFOPDEHALFOP, CA_AUTODEOP, -1,
+    { "HALFOP",    1, CUMODE_h, CA_HALFOP, CA_AUTODEOP, -1,
 	  CHAN_HALFOP_SUCCEEDED, CHAN_HALFOP_FAILED },
-    { "DEHALFOP", 0, CUMODE_h, CA_HALFOPDEHALFOP, CA_AUTOHALFOP, CA_AUTOOP,
+    { "DEHALFOP",  0, CUMODE_h, CA_HALFOP, CA_AUTOHALFOP, CA_AUTOOP,
 	  CHAN_DEHALFOP_SUCCEEDED, CHAN_DEHALFOP_FAILED },
 #endif
+#ifdef HAVE_CHANPROT
+    { "PROTECT",   1, CUMODE_a, CA_PROTECT, -1, -1,
+	  CHAN_PROTECT_SUCCEEDED, CHAN_PROTECT_FAILED },
+    { "DEPROTECT", 0, CUMODE_a, CA_PROTECT, -1, -1,
+	  CHAN_DEPROTECT_SUCCEEDED, CHAN_DEPROTECT_FAILED },
+#endif
 };
 
 static void do_opvoice(User *u, const char *cmd)
@@ -2611,7 +2749,7 @@ static void do_opvoice(User *u, const char *cmd)
     } else if (!(c = findchan(chan))) {
 	notice_lang(s_ChanServ, u, CHAN_X_NOT_IN_USE, chan);
     } else if (c->bouncy_modes) {
-	notice_lang(s_ChanServ, u, CHAN_BOUNCY_MODES, chan, cmd);
+	notice_lang(s_ChanServ, u, CHAN_BOUNCY_MODES, cmd);
     } else if (!(ci = c->ci)) {
 	notice_lang(s_ChanServ, u, CHAN_X_NOT_REGISTERED, chan);
     } else if (ci->flags & CI_VERBOTEN) {
@@ -2656,6 +2794,9 @@ static void do_opvoice(User *u, const char *cmd)
 		   cmd, target, u->nick);
 	}
 	notice_lang(s_ChanServ, u, success_msg, target, chan);
+	/* If it was an OP command, update the last-used time */
+	if (strcmp(cmd, "OP") == 0)
+	    ci->last_used = time(NULL);
     }
 }
 
@@ -2691,6 +2832,18 @@ static void do_dehalfop(User *u)
 }
 #endif  /* HAVE_HALFOP */
 
+#ifdef HAVE_CHANPROT
+static void do_protect(User *u)
+{
+    do_opvoice(u, "PROTECT");
+}
+
+static void do_deprotect(User *u)
+{
+    do_opvoice(u, "DEPROTECT");
+}
+#endif  /* HAVE_CHANPROT */
+
 /*************************************************************************/
 
 static void do_unban(User *u)
@@ -2704,7 +2857,7 @@ static void do_unban(User *u)
     } else if (!(c = findchan(chan))) {
 	notice_lang(s_ChanServ, u, CHAN_X_NOT_IN_USE, chan);
     } else if (c->bouncy_modes) {
-	notice_lang(s_ChanServ, u, CHAN_BOUNCY_MODES, chan, "UNBAN");
+	notice_lang(s_ChanServ, u, CHAN_BOUNCY_MODES, "UNBAN");
     } else if (!(ci = c->ci)) {
 	notice_lang(s_ChanServ, u, CHAN_X_NOT_REGISTERED, chan);
     } else if (ci->flags & CI_VERBOTEN) {
@@ -2738,7 +2891,7 @@ static void do_clear(User *u)
     } else if (!(c = findchan(chan))) {
 	notice_lang(s_ChanServ, u, CHAN_X_NOT_IN_USE, chan);
     } else if (c->bouncy_modes) {
-	notice_lang(s_ChanServ, u, CHAN_BOUNCY_MODES, chan, "CLEAR");
+	notice_lang(s_ChanServ, u, CHAN_BOUNCY_MODES, "CLEAR");
     } else if (!(ci = c->ci)) {
 	notice_lang(s_ChanServ, u, CHAN_X_NOT_REGISTERED, chan);
     } else if (ci->flags & CI_VERBOTEN) {
@@ -2816,7 +2969,7 @@ static void do_forbid(User *u)
     char *chan = strtok(NULL, " ");
 
     /* Assumes that permission checking has already been done. */
-    if (!chan) {
+    if (!chan || *chan != '#') {
 	syntax_error(s_ChanServ, u, "FORBID", CHAN_FORBID_SYNTAX);
 	return;
     }
@@ -2868,6 +3021,7 @@ static void do_suspend(User *u)
     } else if (ci->suspendinfo) {
 	notice_lang(s_ChanServ, u, CHAN_SUSPEND_ALREADY_SUSPENDED, chan);
     } else {
+	Channel *c;
 	if (expiry)
 	    expires = dotime(expiry);
 	else
@@ -2878,9 +3032,13 @@ static void do_suspend(User *u)
 	} else if (expires > 0) {
 	    expires += time(NULL);	/* Set an absolute time */
 	}
-	log("%s: %s SUSPENDED %s", s_ChanServ, u->nick, ci->name);
+	log("%s: %s suspended %s", s_ChanServ, u->nick, ci->name);
 	suspend(ci, reason, u->nick, expires);
 	notice_lang(s_ChanServ, u, CHAN_SUSPEND_SUCCEEDED, chan);
+	c = findchan(chan);
+	if (c)
+	    clear_channel(c, CLEAR_USERS,
+			  "Use of this channel has been forbidden");
 	if (readonly)
 	    notice_lang(s_ChanServ, u, READ_ONLY_MODE);
     }
@@ -2904,7 +3062,7 @@ static void do_unsuspend(User *u)
     } else {
 	if (readonly)
 	    notice_lang(s_ChanServ, u, READ_ONLY_MODE);
-	log("%s: %s UNSUSPENDED %s", s_ChanServ, u->nick, ci->name);
+	log("%s: %s unsuspended %s", s_ChanServ, u->nick, ci->name);
 	unsuspend(ci, 1);
 	notice_lang(s_ChanServ, u, CHAN_UNSUSPEND_SUCCEEDED, chan);
     }
