@@ -61,6 +61,8 @@ int   NoSplitRecovery;
 int   StrictPasswords;
 int   BadPassLimit;
 int   BadPassTimeout;
+int   BadPassWarning;
+int   BadPassSuspend;
 int   UpdateTimeout;
 int   ExpireTimeout;
 int   ReadTimeout;
@@ -75,6 +77,7 @@ int   NSDefFlags;
 int   NSRegDelay;
 int   NSRequireEmail;
 int   NSExpire;
+int   NSExpireWarning;
 int   NSAccessMax;
 char *NSEnforcerUser;
 char *NSEnforcerHost;
@@ -196,7 +199,9 @@ Directive directives[] = {
     { "AutokillDB",       { { PARAM_STRING, 0, &AutokillDBName } } },
     { "AutokillExpiry",   { { PARAM_TIME, 0, &AutokillExpiry } } },
     { "BadPassLimit",     { { PARAM_POSINT, 0, &BadPassLimit } } },
+    { "BadPassSuspend",   { { PARAM_POSINT, 0, &BadPassSuspend } } },
     { "BadPassTimeout",   { { PARAM_TIME, 0, &BadPassTimeout } } },
+    { "BadPassWarning",   { { PARAM_POSINT, 0, &BadPassWarning } } },
     { "ChanServDB",       { { PARAM_STRING, 0, &ChanDBName } } },
     { "ChanServName",     { { PARAM_STRING, 0, &s_ChanServ },
                             { PARAM_STRING, 0, &desc_ChanServ } } },
@@ -267,6 +272,7 @@ Directive directives[] = {
     { "NSDisableLinkCommand",{{PARAM_SET, 0, &NSDisableLinkCommand } } },
     { "NSEnforcerUser",   { { PARAM_STRING, 0, &temp_nsuserhost } } },
     { "NSExpire",         { { PARAM_TIME, 0, &NSExpire } } },
+    { "NSExpireWarning",  { { PARAM_TIME, 0, &NSExpireWarning } } },
     { "NSForceNickChange",{ { PARAM_SET, 0, &NSForceNickChange } } },
     { "NSGuestNickPrefix",{ { PARAM_STRING, 0, &NSGuestNickPrefix } } },
     { "NSListMax",        { { PARAM_POSINT, 0, &NSListMax } } },
