@@ -94,7 +94,7 @@ static int set_group(void)
 
 static int parse_options(int ac, char **av)
 {
-    int i;
+    int i, n;
     char *s, *t;
 
     debug = 0;
@@ -128,14 +128,13 @@ static int parse_options(int ac, char **av)
 		s = av[i];
 		t = strchr(s, ':');
 		if (t) {
-		    *t++ = 0;
-		    if (atoi(t) >= 0)
-			LocalPort = atoi(t);
-		    else {
-			fprintf(stderr, "-local: port number must be a positive integer or 0.  Using default.");
-			return -1;
-		    }
-		    t[-1] = ':';
+		    *t = 0;
+		    n = atoi(t+1);
+		    if (n >= 0)
+			LocalPort = n;
+		    else
+			fprintf(stderr, "-local: port number must be a positive integer or 0.  Using default.\n");
+		    *t = ':';
 		}
 		LocalHost = s;
 	    } else if (strcmp(s, "name") == 0) {
@@ -150,6 +149,21 @@ static int parse_options(int ac, char **av)
 		    return -1;
 		}
 		ServerDesc = av[i];
+	    } else if (strcmp(s, "numeric") == 0) {
+#ifdef IRC_UNREAL
+		if (++i >= ac) {
+		    fprintf(stderr, "-numeric requires a parameter\n");
+		    return -1;
+		}
+		n = atoi(av[i]);
+		if (n < 0 || n > 254)
+		    fprintf(stderr, "-numeric parameter must be between 0 and 254 inclusive.  Using default.\n");
+		else
+		    ServerNumeric = n;
+#else
+		fprintf(stderr, "-numeric available only with Unreal IRC daemon\n");
+		return -1;
+#endif
 	    } else if (strcmp(s, "user") == 0) {
 		if (++i >= ac) {
 		    fprintf(stderr, "-user requires a parameter\n");
@@ -220,8 +234,11 @@ The following options are recognized:
 	-local host[:port]	Connect from the specified address (e.g.
 				    for multihomed servers)
 	-name servername	Our server name (e.g. services.some.net)
-	-desc string		Description of us (e.g. SomeNet Services)
-	-user username		Username for Services' nicks (e.g. services)
+	-desc string		Description of us (e.g. SomeNet Services)\n"
+#ifdef IRC_UNREAL
+"	-numeric numeric	Services server numeric (1-254 or 0 for none)\n"
+#endif
+"	-user username		Username for Services' nicks (e.g. services)
 	-host hostname		Hostname for Services' nicks (e.g. some.net)
 	-dir directory		Directory containing Services' data files
 				    (e.g. /usr/local/lib/services)
