@@ -162,6 +162,8 @@ static void do_listchans(int count_only, const char *chan)
 	    if (ci->mlock_limit)
 		printf(" %d", ci->mlock_limit);
 	    printf("\n");
+	    if (ci->flags & CI_NOEXPIRE)
+		printf("This channel will not expire.\n");
 	}
 
     } else {
