@@ -179,16 +179,20 @@ char *stristr(char *s1, char *s2)
 char *strupper(char *s)
 {
     char *t = s;
-    while (*t)
-	*t++ = toupper(*t);
+    while (*t) {
+	*t = toupper(*t);
+	t++;
+    }
     return s;
 }
 
 char *strlower(char *s)
 {
     char *t = s;
-    while (*t)
-	*t++ = tolower(*t);
+    while (*t) {
+	*t = tolower(*t);
+	t++;
+    }
     return s;
 }
 
