@@ -73,6 +73,7 @@ static void link_fix_users(void);
 static void suspend(NickInfo *ni, const char *reason,
 		    const char *who, const time_t expires);
 static void unsuspend(NickInfo *ni, int set_time);
+static void nick_bad_password(User *u, NickInfo *ni);
 
 static void collide(NickInfo *ni, int from_timeout);
 static void release(NickInfo *ni, int from_timeout);
@@ -119,6 +120,7 @@ static Command cmds[] = {
     { "HELP",     do_help,     NULL,  -1,                     -1,-1,-1,-1 },
     { "REGISTER", do_register, NULL,  NICK_HELP_REGISTER,     -1,-1,-1,-1 },
     { "IDENTIFY", do_identify, NULL,  NICK_HELP_IDENTIFY,     -1,-1,-1,-1 },
+    { "SIDENTIFY",do_identify, NULL,  -1,                     -1,-1,-1,-1 },
     { "DROP",     do_drop,     NULL,  -1,
 		NICK_HELP_DROP, NICK_SERVADMIN_HELP_DROP,
 		NICK_SERVADMIN_HELP_DROP, NICK_SERVADMIN_HELP_DROP },
@@ -403,28 +405,44 @@ int validate_user(User *u)
 	/*
 	 * This can be exploited to gain improper privilege if an attacker
 	 * has the same Services stamp, username and hostname as the
-	 * victim.  Under ircd.dal 4.4.15+ and other servers supporting a
-	 * Services stamp, Services guarantees that the first condition
-	 * cannot occur unless the stamp counter rolls over (2^32 client
-	 * connections).  Under other servers, an attack is theoretically
-	 * possible, but would require access to the computer the victim
-	 * is using for IRC or the DNS servers for the victim's domain and
-	 * IP address range in order to have the same hostname, and would
-	 * require that the attacker connect so that he has the same
-	 * server timestamp as the victim (this could be accomplished by
-	 * finding a server with a clock slower than that of the victim's
-	 * server and timing the connection attempt properly).
+	 * victim.
+	 *
+	 * Under ircd.dal 4.4.15+ and other servers supporting a Services
+	 * stamp, Services guarantees that the first condition cannot occur
+	 * unless the stamp counter rolls over (2^32-1 client connections).
+	 * This is practically infeasible.  As an example, on a network of
+	 * 30 servers, an attack introducing 10 new clients every second on
+	 * every server would need to be sustained for 166 days (5.5 months)
+	 * to cause the stamp to roll over.
+	 *
+	 * Under other servers, an attack is theoretically possible, but
+	 * would require access to the computer the victim is using for IRC
+	 * or the DNS servers for the victim's domain and IP address range
+	 * in order to have the same hostname, and would require that the
+	 * attacker connect so that he has the same server timestamp as the
+	 * victim (this could be accomplished by finding a server with a
+	 * clock slower than that of the victim's server and timing the
+	 * connection attempt properly).
+	 *
+	 * When using Unreal, the username and hostname are not checked,
+	 * since the real hostname is not saved, only the fake one.
+	 * However, since Unreal supports Services stamps, this is still
+	 * believed safe (see above).
 	 *
 	 * If someone gets a hacked server into your network, all bets are
 	 * off.
 	 */
 	if (ni->id_stamp != 0 && u->services_stamp == ni->id_stamp) {
+#ifndef IRC_UNREAL
 	    char buf[BUFSIZE];
 	    snprintf(buf, sizeof(buf), "%s@%s", u->username, u->host);
 	    if (strcmp(buf, ni->last_usermask) == 0) {
+#endif
 		ni->status |= NS_IDENTIFIED;
 		return 1;
+#ifndef IRC_UNREAL
 	    }
+#endif
 	}
     }
 
@@ -435,8 +453,13 @@ int validate_user(User *u)
 	ni->last_seen = time(NULL);
 	if (ni->last_usermask)
 	    free(ni->last_usermask);
+#ifdef IRC_UNREAL
+	ni->last_usermask = smalloc(strlen(u->username)+strlen(u->fakehost)+2);
+	sprintf(ni->last_usermask, "%s@%s", u->username, u->fakehost);
+#else
 	ni->last_usermask = smalloc(strlen(u->username)+strlen(u->host)+2);
 	sprintf(ni->last_usermask, "%s@%s", u->username, u->host);
+#endif
 	if (ni->last_realname)
 	    free(ni->last_realname);
 	ni->last_realname = sstrdup(u->realname);
@@ -453,6 +476,7 @@ int validate_user(User *u)
     if ((u->ni->flags & NI_KILLPROTECT) && !on_access) {
 	if (u->ni->flags & NI_KILL_IMMED) {
 	    collide(ni, 0);
+	    return 0;
 	} else if (u->ni->flags & NI_KILL_QUICK) {
 #ifdef HAVE_NICKCHANGE
 	    if (NSForceNickChange)
@@ -472,6 +496,27 @@ int validate_user(User *u)
 	}
     }
 
+    if (NSExpire && NSExpireWarning && !(ni->status & NS_NOEXPIRE)) {
+	int time_left = NSExpire - (time(NULL) - ni->last_seen);
+	if (time_left <= NSExpireWarning) {
+	    int msg;
+	    if (time_left >= 86400) {
+		time_left /= 86400;
+		msg = (time_left==1 ? STR_DAY : STR_DAYS);
+	    } else if (time_left >= 3600) {
+		time_left /= 3600;
+		msg = (time_left==1 ? STR_HOUR : STR_HOURS);
+	    } else {
+		time_left /= 60;
+		if (time_left < 1)
+		    time_left = 1;
+		msg = (time_left==1 ? STR_MINUTE : STR_MINUTES);
+	    }
+	    notice_lang(s_NickServ, u, NICK_EXPIRES_SOON, time_left,
+			getstring(ni,msg), s_NickServ, s_NickServ);
+	}
+    }
+
     return 0;
 }
 
@@ -537,10 +582,6 @@ void expire_nicks()
     NickInfo *ni, *next;
     time_t now = time(NULL);
 
-    /* Assumption: this routine takes less than NSExpire seconds to run.
-     * If it doesn't, some users may end up with invalid user->ni pointers.
-     * (FIXME: why? --AC) */
-
     for (u = firstuser(); u; u = nextuser()) {
 	if (u->real_ni && (u->real_ni->status & (NS_IDENTIFIED|NS_RECOGNIZED))){
 	    if (debug >= 2)
@@ -558,6 +599,11 @@ void expire_nicks()
 	 && !ni->suspendinfo
 	) {
 	    log("%s: Expiring nickname %s", s_NickServ, ni->nick);
+	    u = finduser(ni->nick);
+	    if (u) {
+		notice_lang(s_NickServ, u, NICK_EXPIRED);
+		u->ni = u->real_ni = NULL;
+	    }
 	    delnick(ni);
 	} else if (ni->suspendinfo && ni->suspendinfo->expires
 		   && now >= ni->suspendinfo->expires) {
@@ -659,6 +705,8 @@ static int delnick(NickInfo *ni)
 	free(ni->last_usermask);
     if (ni->last_realname)
 	free(ni->last_realname);
+    if (ni->last_quit)
+	free(ni->last_quit);
     if (ni->access) {
 	for (i = 0; i < ni->accesscount; i++) {
 	    if (ni->access[i])
@@ -814,6 +862,31 @@ static void unsuspend(NickInfo *ni, int set_time)
     }
 }
 
+/*************************************************************************/
+
+/* Register a bad password attempt for a nickname. */
+
+static void nick_bad_password(User *u, NickInfo *ni)
+{
+    bad_password(s_NickServ, u, ni->nick);
+    ni->bad_passwords++;
+    if (BadPassWarning && ni->bad_passwords == BadPassWarning) {
+	wallops(s_NickServ, "\2Warning:\2 Repeated bad password attempts"
+	                    " for nick %s", ni->nick);
+    }
+    if (BadPassSuspend && ni->bad_passwords == BadPassSuspend) {
+	time_t expire = 0;
+	if (NSSuspendExpire)
+	    expire = time(NULL) + NSSuspendExpire;
+	suspend(ni, "Too many bad passwords (automatic suspend)",
+		s_NickServ, expire);
+	log("%s: Automatic suspend for %s (too many bad passwords)",
+	    s_NickServ, ni->nick);
+	/* Clear bad password count for when nick is unsuspended */
+	ni->bad_passwords = 0;
+    }
+}
+
 /*************************************************************************/
 /*************************************************************************/
 
@@ -1078,12 +1151,12 @@ static void do_register(User *u)
 
     } else if (u->real_ni) {	/* i.e. there's already such a nick regged */
 	if (u->real_ni->status & NS_VERBOTEN) {
-	    log("%s: %s@%s tried to register FORBIDden nick %s", s_NickServ,
+	    log("%s: %s@%s tried to register forbidden nick %s", s_NickServ,
 			u->username, u->host, u->nick);
 	    notice_lang(s_NickServ, u, NICK_CANNOT_BE_REGISTERED, u->nick);
 	} else {
 	    if (u->ni->suspendinfo)
-		log("%s: %s@%s tried to register SUSPENDed nick %s",
+		log("%s: %s@%s tried to register suspended nick %s",
 		    s_NickServ, u->username, u->host, u->nick);
 	    notice_lang(s_NickServ, u, NICK_ALREADY_REGISTERED, u->nick);
 	}
@@ -1133,20 +1206,30 @@ static void do_register(User *u)
 	ni->memos.memomax = MSMaxMemos;
 	ni->channelcount = 0;
 	ni->channelmax = CSMaxReg;
+#ifdef IRC_UNREAL
+	ni->last_usermask = smalloc(strlen(u->username)+strlen(u->fakehost)+2);
+	sprintf(ni->last_usermask, "%s@%s", u->username, u->fakehost);
+#else
 	ni->last_usermask = smalloc(strlen(u->username)+strlen(u->host)+2);
 	sprintf(ni->last_usermask, "%s@%s", u->username, u->host);
+#endif
 	ni->last_realname = sstrdup(u->realname);
 	ni->time_registered = ni->last_seen = time(NULL);
 	ni->accesscount = 1;
 	ni->access = smalloc(sizeof(char *));
-	ni->access[0] = create_mask(u);
+	ni->access[0] = create_mask(u, 0);
 	ni->language = DEF_LANGUAGE;
 	ni->link = NULL;
 	if (email)
 	    ni->email = sstrdup(email);
 	u->ni = u->real_ni = ni;
-	log("%s: `%s' registered by %s@%s", s_NickServ,
-	    u->nick, u->username, u->host);
+	if (email) {
+	    log("%s: `%s' registered by %s@%s (%s)", s_NickServ,
+		u->nick, u->username, u->host, email);
+	} else {
+	    log("%s: `%s' registered by %s@%s", s_NickServ,
+		u->nick, u->username, u->host);
+	}
 	notice_lang(s_NickServ, u, NICK_REGISTERED, u->nick, ni->access[0]);
 #ifndef USE_ENCRYPTION
 	notice_lang(s_NickServ, u, NICK_PASSWORD_IS, ni->pass);
@@ -1180,25 +1263,31 @@ static void do_identify(User *u)
     } else if (!(res = check_password(pass, ni->pass))) {
 	log("%s: Failed IDENTIFY for %s!%s@%s",
 		s_NickServ, u->nick, u->username, u->host);
-	bad_password(s_NickServ, u, ni->nick);
+	nick_bad_password(u, ni);
 
     } else if (res == -1) {
 	notice_lang(s_NickServ, u, NICK_IDENTIFY_FAILED);
 
     } else {
+	ni->bad_passwords = 0;
 	ni->status |= NS_IDENTIFIED;
 	ni->id_stamp = u->services_stamp;
 	ni->last_seen = time(NULL);
 	if (!(ni->status & NS_RECOGNIZED)) {
 	    if (ni->last_usermask)
 		free(ni->last_usermask);
+#ifdef IRC_UNREAL
+	    ni->last_usermask = smalloc(strlen(u->username)+strlen(u->fakehost)+2);
+	    sprintf(ni->last_usermask, "%s@%s", u->username, u->fakehost);
+#else
 	    ni->last_usermask = smalloc(strlen(u->username)+strlen(u->host)+2);
 	    sprintf(ni->last_usermask, "%s@%s", u->username, u->host);
+#endif
 	    if (ni->last_realname)
 		free(ni->last_realname);
 	    ni->last_realname = sstrdup(u->realname);
 	}
-#ifdef HAVE_NICKCHANGE
+#ifdef IRC_DAL4_4_15
 	if (nick_is_services_admin(ni))
 	    send_cmd(s_NickServ, "SVSMODE %s :+ra", u->nick);
 	else
@@ -1207,7 +1296,7 @@ static void do_identify(User *u)
 	    send_cmd(s_NickServ, "SVSMODE %s :+%s", u->nick,
 		     mode_flags_to_string(UMODE_REG, MODE_USER));
 	log("%s: %s!%s@%s identified for nick %s", s_NickServ,
-			u->nick, u->username, u->host, u->nick);
+	    u->nick, u->username, u->host, u->nick);
 	notice_lang(s_NickServ, u, NICK_IDENTIFY_SUCCEEDED);
 	if (!(ni->status & NS_RECOGNIZED)) {
 	    check_memos(u);
@@ -1484,10 +1573,14 @@ static void do_set_email(User *u, NickInfo *ni, char *param)
 	    notice_lang(s_NickServ, u, BAD_EMAIL);
 	} else {
 	    ni->email = sstrdup(param);
+	    log("%s: %s E-mail address changed to %s by %s!%s@%s",
+		s_NickServ, ni->nick, param, u->nick, u->username, u->host);
 	    notice_lang(s_NickServ, u, NICK_SET_EMAIL_CHANGED, ni->nick,param);
 	}
     } else {
 	ni->email = NULL;
+	log("%s: %s E-mail address cleared by %s!%s@%s",
+	    s_NickServ, ni->nick, u->nick, u->username, u->host);
 	notice_lang(s_NickServ, u, NICK_UNSET_EMAIL, ni->nick);
     }
 }
@@ -1648,7 +1741,7 @@ static void do_access(User *u)
 	    return;
 	}
 	for (access = ni->access, i = 0; i < ni->accesscount; access++, i++) {
-	    if (strcmp(*access, mask) == 0) {
+	    if (stricmp(*access, mask) == 0) {
 		notice_lang(s_NickServ, u,
 			NICK_ACCESS_ALREADY_PRESENT, *access);
 		return;
@@ -1734,7 +1827,7 @@ static void do_link(User *u)
     } else if (!(res = check_password(pass, target->pass))) {
 	log("%s: LINK: bad password for %s by %s!%s@%s",
 		s_NickServ, nick, u->nick, u->username, u->host);
-	bad_password(s_NickServ, u, target->nick);
+	nick_bad_password(u, target);
 
     } else if (res == -1) {
 	notice_lang(s_NickServ, u, NICK_LINK_FAILED);
@@ -1743,6 +1836,8 @@ static void do_link(User *u)
 	NickInfo *tmp, *top;
 	int depth;
 
+	target->bad_passwords = 0;
+
 	/* Get the top-level nick.  This also has the side effect of
 	 * ensuring we're not about to run into an infinite loop
 	 * (getlink() will check that for us). */
@@ -1853,10 +1948,11 @@ static void do_unlink(User *u)
 				!(res = check_password(pass, ni->pass))) {
 	    log("%s: UNLINK: bad password for %s by %s!%s@%s",
 		s_NickServ, nick, u->nick, u->username, u->host);
-	    bad_password(s_NickServ, u, ni->nick);
+	    nick_bad_password(u, ni);
 	} else if (res == -1) {
 	    notice_lang(s_NickServ, u, NICK_UNLINK_FAILED);
 	} else {
+	    ni->bad_passwords = 0;
 	    linkname = ni->link->nick;
 	    delink(ni);
 	    link_fix_users();
@@ -2196,13 +2292,14 @@ static void do_recover(User *u)
     } else if (pass) {
 	int res = check_password(pass, ni->pass);
 	if (res == 1) {
+	    ni->bad_passwords = 0;
 	    collide(ni, 0);
 	    notice_lang(s_NickServ, u, NICK_RECOVERED, s_NickServ, nick);
 	} else {
 	    if (res == 0) {
 		log("%s: RECOVER: invalid password for %s by %s!%s@%s",
 		    s_NickServ, nick, u->nick, u->username, u->host);
-		bad_password(s_NickServ, u, ni->nick);
+		nick_bad_password(u, ni);
 	    } else { /* res == -1 */
 		log("%s: RECOVER: check_password failed for %s",
 		    s_NickServ, nick);
@@ -2236,15 +2333,16 @@ static void do_release(User *u)
     } else if (pass) {
 	int res = check_password(pass, ni->pass);
 	if (res == 1) {
+	    ni->bad_passwords = 0;
 	    release(ni, 0);
 	    notice_lang(s_NickServ, u, NICK_RELEASED);
 	} else {
 	    if (res == 0) {
 		log("%s: RELEASE: invalid password for %s by %s!%s@%s",
 		    s_NickServ, nick, u->nick, u->username, u->host);
-		bad_password(s_NickServ, u, ni->nick);
+		nick_bad_password(u, ni);
 	    } else { /* res == -1 */
-		log("%s: RECOVER: check_password failed for %s",
+		log("%s: RELEASE: check_password failed for %s",
 		    s_NickServ, nick);
 		notice_lang(s_NickServ, u, ACCESS_DENIED);
 	    }
@@ -2282,16 +2380,17 @@ static void do_ghost(User *u)
 	int res = check_password(pass, ni->pass);
 	if (res == 1) {
 	    char buf[NICKMAX+32];
+	    ni->bad_passwords = 0;
 	    snprintf(buf, sizeof(buf), "GHOST command used by %s", u->nick);
 	    kill_user(s_NickServ, nick, buf);
 	    notice_lang(s_NickServ, u, NICK_GHOST_KILLED, nick);
 	} else {
 	    if (res == 0) {
-		log("%s: RELEASE: invalid password for %s by %s!%s@%s",
+		log("%s: GHOST: invalid password for %s by %s!%s@%s",
 		    s_NickServ, nick, u->nick, u->username, u->host);
-		bad_password(s_NickServ, u, ni->nick);
+		nick_bad_password(u, ni);
 	    } else { /* res == -1 */
-		log("%s: RECOVER: check_password failed for %s",
+		log("%s: GHOST: check_password failed for %s",
 		    s_NickServ, nick);
 		notice_lang(s_NickServ, u, ACCESS_DENIED);
 	    }
@@ -2493,7 +2592,7 @@ static void do_listnick(User *u)
     strscpy(buf2, ni->email ? ni->email : "-", sizeof(buf2));
     if (ni->suspendinfo) {
 	SuspendInfo *si = ni->suspendinfo;
-	snprintf(buf3, sizeof(buf3), "%s.%d.%d.%s",
+	snprintf(buf3, sizeof(buf3), "%s.%ld.%ld.%s",
 		 si->who, si->suspended, si->expires,
 		 si->reason ? si->reason : "-");
 	strnrepl(buf3, sizeof(buf3), " ", "_");
