@@ -188,7 +188,19 @@ void operserv(const char *source, char *buf)
 	return;
     }
 
-    log("%s: %s: %s", s_OperServ, source, buf);
+    /* Don't log stuff that might be passwords */
+    if (strnicmp(buf, "SU ", 3) == 0) {
+	log("%s: %s: SU xxxxxx", s_OperServ, source);
+    } else if (strnicmp(buf, "SET ", 4) == 0
+	       && (s = stristr(buf, "SUPASS")) != NULL
+	       && strspn(buf+4, " ") == s-(buf+4)) {
+	/* All that was needed to make sure someone doesn't fool us with
+	 * "SET READONLY ON SUPASS".  Which wouldn't work anyway, but we
+	 * ought to log it properly... */
+	log("%s: %s: SET SUPASS xxxxxx", s_OperServ, source);
+    } else {
+	log("%s: %s: %s", s_OperServ, source, buf);
+    }
 
     cmd = strtok(buf, " ");
     if (!cmd) {
@@ -689,6 +701,9 @@ static void do_stats(User *u)
     struct tm *tm;
     char timebuf[64];
 
+    if (extra && stricmp(extra, "UPTIME") == 0)
+	extra = NULL;
+
     if (extra && stricmp(extra, "ALL") != 0) {
 	if (stricmp(extra, "AKILL") == 0) {
 	    int timeout = AutokillExpiry+59;
@@ -710,7 +725,6 @@ static void do_stats(User *u)
 		notice_lang(s_OperServ, u, OPER_STATS_AKILL_EXPIRE_MIN);
 	    else
 		notice_lang(s_OperServ, u, OPER_STATS_AKILL_EXPIRE_NONE);
-	    return;
 	} else if (stricmp(extra, "RESET") == 0) {
 	    maxusercnt = usercnt;
 	    maxusertime = time(NULL);
@@ -719,6 +733,7 @@ static void do_stats(User *u)
 	    notice_lang(s_OperServ, u, OPER_STATS_UNKNOWN_OPTION,
 			strupper(extra));
 	}
+	return;
     }
 
     notice_lang(s_OperServ, u, OPER_STATS_CURRENT_USERS, usercnt, opcnt);
@@ -1069,8 +1084,9 @@ static void do_su(User *u)
     int res;
 
     if (!is_services_admin(u)) {
-	wallops(s_OperServ, "NOTICE: %s!%s@%s attempted to use SU command "
-		"(not Services admin)", u->nick, u->username, u->host);
+	wallops(s_OperServ, "\2NOTICE:\2 %s!%s@%s attempted to use SU "
+		"command (not Services admin)",
+		u->nick, u->username, u->host);
 	notice_lang(s_OperServ, u, PERMISSION_DENIED);
 	return;
     }
@@ -1086,7 +1102,7 @@ static void do_su(User *u)
     } else if (res == 0) {
 	log("%s: Failed SU by %s!%s@%s",
 	    s_OperServ, u->nick, u->username, u->host);
-	wallops(s_OperServ, "NOTICE: Failed SU by %s!%s@%s",
+	wallops(s_OperServ, "\2NOTICE:\2 Failed SU by %s!%s@%s",
 		u->nick, u->username, u->host);
 	bad_password(s_OperServ, u, "Services root");
     } else {
