@@ -224,8 +224,6 @@ static int flush_write_buffer(int wait)
 	int maxwrite, nwritten;
 	if (write_curpos > write_bufend)	/* wrapped around? */
 	    maxwrite = write_buftop - write_curpos;
-	else if (write_bufend == write_netbuf)
-	    maxwrite = write_buftop - write_curpos - 1;
 	else
 	    maxwrite = write_bufend - write_curpos;
 	nwritten = write(write_fd, write_curpos, maxwrite);
