@@ -22,15 +22,16 @@ static int checking_timeouts = 0;
 void send_timeout_list(User *u)
 {
     Timeout *to, *last;
+    uint32 now = time_msec();
 
-    notice(s_OperServ, u->nick, "Now: %ld", time(NULL));
+    notice(s_OperServ, u->nick, "Now: %u.%03u", now/1000, now%1000);
     for (to = timeouts, last = NULL; to; last = to, to = to->next) {
-	notice(s_OperServ, u->nick, "%p: %u: %p (%p)",
-	       to, to->timeout, to->code, to->data);
+	notice(s_OperServ, u->nick, "%p: %u.%03u: %p (%p)",
+	       to, to->timeout/1000, to->timeout%1000, to->code, to->data);
 	if (to->prev != last)
 	    notice(s_OperServ, u->nick,
-			"    to->prev incorrect!  expected=%p seen=%p",
-			last, to->prev);
+		   "    to->prev incorrect!  expected=%p seen=%p",
+		   last, to->prev);
     }
 }
 
@@ -48,8 +49,10 @@ void check_timeouts(void)
     if (checking_timeouts)
 	fatal("check_timeouts() called recursively!");
     checking_timeouts = 1;
-    if (debug >= 2)
-	log("debug: Checking timeouts at time_msec = %u", now);
+    if (debug >= 2) {
+	log("debug: Checking timeouts at time_msec = %u.%03u",
+	    now/1000, now%1000);
+    }
 
     to = timeouts;
     while (to) {
@@ -105,8 +108,8 @@ Timeout *add_timeout(int delay, void (*code)(Timeout *), int repeat)
 Timeout *add_timeout_ms(uint32 delay, void (*code)(Timeout *), int repeat)
 {
     Timeout *t = smalloc(sizeof(Timeout));
-    t->settime = time_msec();
-    t->timeout = t->settime + delay;
+    t->settime = time(NULL);
+    t->timeout = time_msec() + delay;
     t->code = code;
     t->repeat = repeat;
     t->next = timeouts;
