@@ -17,7 +17,8 @@
 typedef struct timeout_ Timeout;
 struct timeout_ {
     Timeout *next, *prev;
-    uint32 settime, timeout;	/* In milliseconds (time_msec()) */
+    time_t settime;		/* Time timer was set (from time()) */
+    uint32 timeout;		/* In milliseconds (time_msec()) */
     int repeat;			/* Does this timeout repeat indefinitely? */
     void (*code)(Timeout *);	/* This structure is passed to the code */
     void *data;			/* Can be anything */
