@@ -347,7 +347,10 @@ int do_nick(const char *source, int ac, char **av)
     if (!*source) {
 	/* This is a new user; create a User structure for it. */
 #ifdef IRC_DAL4_4_15
-	static uint32 servstamp = 0;
+	/* Unreal (and other ircds in this class?) upper-bound values at
+	 * 2^31-1, so we limit ourselves to 31 bits even though our field
+	 * (and the field in ircd/include/struct.h) is unsigned... */
+	static int32 servstamp = 0;
 	int i;
 #endif
 
@@ -386,12 +389,12 @@ int do_nick(const char *source, int ac, char **av)
 	    user->services_stamp = i;
 	} else {
 	    if (servstamp == 0) {
-		servstamp = rand() + rand();
+		servstamp = rand() & 0x7FFFFFFF;
 		if (servstamp == 0)
 		    servstamp++;
 	    }
 	    user->services_stamp = servstamp++;
-	    if (servstamp == 0)
+	    if (servstamp <= 0)
 		servstamp = 1;
 	    send_cmd(ServerName, "SVSMODE %s +d %u", user->nick,
 		     user->services_stamp);
@@ -440,10 +443,17 @@ int do_nick(const char *source, int ac, char **av)
 	new_ni = findnick(av[0]);
 	if (new_ni)
 	    new_ni = getlink(new_ni);
-	if (new_ni != user->ni)
+	if (new_ni != user->ni) {
 	    cancel_user(user);
-	else if (new_ni)
-	    ni_changed = 0;
+	} else if (new_ni) {
+	    if (!user->real_ni) {
+		log("user: do_nick BUG: user->ni set but real_ni not set!");
+	    } else if (user->real_ni->status & (NS_IDENTIFIED|NS_RECOGNIZED)) {
+		ni_changed = 0;
+	    } else {
+		cancel_user(user);
+	    }
+	}
 	change_user_nick(user, av[0]);
     }
 
@@ -583,13 +593,24 @@ void do_sjoin(const char *source, int ac, char **av)
         t = nick + strcspn(nick, " ");
 	if (*t)
 	    *t++ = 0;
+#ifdef IRC_UNREAL
+	if (*nick == '&' || *nick == '"') {
+	    /* Ban (&) or exception (") */
+	    char *av[3];
+	    av[0] = channel;
+	    av[1] = (*nick=='&') ? "+b" : "+e";
+	    av[2] = nick+1;
+	    do_cmode(source, 3, av);
+	    continue;
+	}
+#endif
 	do {
 	    thismode = cumode_prefix_to_flag(*nick);
 	    modes |= thismode;
 	} while (thismode && nick++); /* increment nick only if thismode!=0 */
 	user = finduser(nick);
 	if (!user) {
-	    log("channel: SJOIN to channel %s for non-existant nick %s (%s)",
+	    log("channel: SJOIN to channel %s for non-existent nick %s (%s)",
 		channel, nick, merge_args(ac-1, av));
 	    continue;
 	}
@@ -736,9 +757,10 @@ void do_umode(const char *source, int ac, char **av)
 	return;
     }
 #ifdef IRC_DAL4_4_15
-    if (strchr(av[1], 'd')) {
+    if (strchr(av[1], 'd') && ac > 2) {
 	log("user: MODE tried to change services stamp: %s",
 	    merge_args(ac, av));
+	send_cmd(s_NickServ, "SVSMODE %s +d %u", av[0], user->services_stamp);
     }
 #endif
     if (debug)
@@ -921,14 +943,14 @@ int is_voiced(const char *nick, const char *chan)
 /*************************************************************************/
 
 /* Does the user's usermask match the given mask (either nick!user@host or
- * just user@host)?
+ * just user@host)?  Also checks fakehost where supported.
  */
 
 int match_usermask(const char *mask, User *user)
 {
     char *mask2 = sstrdup(mask);
     char *nick, *username, *host;
-    int result;
+    int match_user, match_host, result;
 
     if (strchr(mask2, '!')) {
 	nick = strtok(mask2, "!");
@@ -942,13 +964,17 @@ int match_usermask(const char *mask, User *user)
 	free(mask2);
 	return 0;
     }
+    match_user = match_wild(username, user->username);
+    match_host = match_wild_nocase(host, user->host);
+#ifdef IRC_UNREAL
+    if (user->fakehost)
+	match_host |= match_wild_nocase(host, user->fakehost);
+#endif
     if (nick) {
 	result = match_wild_nocase(nick, user->nick) &&
-		 match_wild(username, user->username) &&
-		 match_wild_nocase(host, user->host);
+		 match_user && match_host;
     } else {
-	result = match_wild(username, user->username) &&
-		 match_wild_nocase(host, user->host);
+	result = match_user && match_host;
     }
     free(mask2);
     return result;
@@ -994,37 +1020,45 @@ void split_usermask(const char *mask, char **nick, char **user, char **host)
  * for named addresses, wildcards the leftmost part of the name unless the
  * name only contains two parts.  The returned character string is malloc'd
  * and should be free'd when done with.
+ *
+ * Where supported, uses the fake host instead of the real one if
+ * use_fakehost is nonzero.
  */
 
-char *create_mask(User *u)
+char *create_mask(User *u, int use_fakehost)
 {
-    char *mask, *s, *end;
+    char *mask, *s, *end, *host;
 
+    host = u->host;
+#ifdef IRC_UNREAL
+    if (use_fakehost && u->fakehost)
+	host = u->fakehost;
+#endif
     /* Get us a buffer the size of the username plus hostname.  The result
      * will never be longer than this (and will often be shorter), thus we
      * can use strcpy() and sprintf() safely.
      */
-    end = mask = smalloc(strlen(u->username) + strlen(u->host) + 2);
+    end = mask = smalloc(strlen(u->username) + strlen(host) + 2);
     end += sprintf(end, "%s@", u->username);
-    if (strspn(u->host, "0123456789.") == strlen(u->host)
+    if (strspn(u->host, "0123456789.") == strlen(host)
 		&& (s = strchr(u->host, '.'))
 		&& (s = strchr(s+1, '.'))
 		&& (s = strchr(s+1, '.'))
 		&& (   !strchr(s+1, '.'))) {	/* IP addr */
-	s = sstrdup(u->host);
+	s = sstrdup(host);
 	*strrchr(s, '.') = 0;
-	if (atoi(u->host) < 192)
+	if (atoi(host) < 192)
 	    *strrchr(s, '.') = 0;
-	if (atoi(u->host) < 128)
+	if (atoi(host) < 128)
 	    *strrchr(s, '.') = 0;
 	sprintf(end, "%s.*", s);
 	free(s);
     } else {
-	if ((s = strchr(u->host+1, '.')) && strchr(s+1, '.')) {
+	if ((s = strchr(host+1, '.')) && strchr(s+1, '.')) {
 	    s = sstrdup(s-1);
 	    *s = '*';
 	} else {
-	    s = sstrdup(u->host);
+	    s = sstrdup(host);
 	}
 	strcpy(end, s);
 	free(s);
